﻿/**
 * 消息端点。
 *
 * 此控制器处理点对点，组及消息消息。为三类消息提供发送及查询功能。
 */
var http = require('http');
var getui = require('getui');
var log = require('../util/log.js');
var objectUtil = require('../util/objectUtil');

var express = require('express');
var router = express.Router();

var systemMsg = require("../models/msg.system");
var privateMsg = require('../models/msg.private');
var groupMsg = require("../models/msg.group");
var notifyMsg = require("../models/msg.notify");
var msgStats = require("../models/msg.stat");
var user = require("../models/user");
var group = require("../models/group");

var APIv1 = require('../include/endpoints').APIv1;
var CHANNELS = require('../include/commons').CHANNELS;
var PLATFORMS = require('../include/commons').PLATFORM;

var MAX_INT = 9007199254740992;

//--------------------------------------------------------------//
//----------------------------消息发送----------------------------//
//--------------------------------------------------------------//

/**
 * 发送System消息。
 *
 * 消息格式：
 *  {
 *      to: "Rose",
 *      title: "System Message",
 *      summary: "You have new job",
 *      contentType: "1",
 *      content: "The patient has been followed in the scheduler, please make new follow plan as soon as possible."
 *  }
 *
 * @param message
 */
router.post(APIv1.Chats.SM, function (req, res) {
    // 检查消息体及消息格式是否正确
    var message = req.body;
    if (!objectUtil.isJsonObject(message)) {
        throw {httpStatus: 406, message: 'Problems parsing JSON.'}
    }

    // 字段判断
    if (message.to == null || message.title == null) {
        throw {httpStatus: 406, message: 'Missing fields.'}
    }

    // 消息处理
    user.getUserStatus(message.to, function (err, userStatus) {
        if (err) {
            console.log("Lookup system message receiver failed: ", err);
            
            throw {httpStatus: 500, message: "Lookup system message receiver failed: "};
        }

        if(userStatus.length == 0){
            throw {httpStatus: 404, message: "User not found."};
        }

        res.writeHead(200);

        var pushable = userStatus.length > 0 && userStatus[0].is_online;
        var notifyMessage = JSON.stringify({type: message.contentType, content: message.content});

        // 保存该条推送信息
        systemMsg.save(message.to,
            message.contentType,
            message.title,
            message.summary,
            message.content,
            function (err, result) {
                if (err) {
                    log.error("Save system message failed, ", err);

                    throw {httpStatus: 500, message: "Save system message failed."};
                } else {
                    // 保存通知到数据库中
                    notifyMessage.saveMessage(message.to,
                        message.contentType,
                        message.title,
                        message.summary,
                        notifyMessage,
                        pushable,
                        function (err, result) {
                            if (err) {
                                log.error("Save system message failed, ", err);

                                throw {httpStatus: 500, message: "Save system notify message failed."};
                            } else {
                                if (pushable == true) {
                                    if (userStatus.platform == PLATFORMS.iOS) {
                                        getui.pushAPN(message.to,
                                            userStatus.token,
                                            message.contentType,
                                            message.title,
                                            message.content,
                                            notifyMessage,
                                            function (err, result) {
                                                err != null ? console.log(err) : console.log(result);
                                            });
                                    } else if (PLATFORMS.iOS.Android) {
                                        getui.pushAndroid(userStatus.client_id,
                                            message.contentType,
                                            message.title,
                                            message.content,
                                            notifyMessage,
                                            userStatus.status,
                                            function (err, result) {
                                                err != null ? console.log(err) : console.log(result);
                                            });
                                    }
                                }
                            }
                        });
                }
            });
    });

    res.writeHead(200);
});

/**
 * 处理Private消息。处理流程分：
 * 1 解析消息，并保存到数据库
 * 2 更新消息统计数据
 * 3 获取目标的状态并构建通知消息，如果用户在线就推送通知消息
 *
 * 消息格式：
 *  {
 *      from: sand,
 *      to: Rose,
 *      contentType: "1",
 *      content: "Please follow the patient as soon as possible."
 *  }
 *
 * @param message
 */
router.post(APIv1.Chats.PM, function (req, res) {
    // 检查消息体及消息格式是否正确
    var message = req.body;
    if (!objectUtil.isJsonObject(message)) {
        throw {httpStatus: 406, message: 'Problems parsing JSON.'}
    }

    // 字段判断
    if (message.from == null || message.to == null || message.contentType == null || message.content == null) {

        throw {httpStatus: 406, message: 'Missing fields.'}
    }

    // 处理消息
    user.isExist(message.to, function (err, result) {
        if (err) {
            console.log('Lookup receiving users failed: ', err);

            throw {httpStatus: 500, message: 'Lookup receiving users failed.'};
        }

        if (result.length == 0) {
            throw {httpStatus: 404, message: 'Receiving users not found.'};
        }

        // 保存消息
        privateMsg.save(message.to, message.from, message.contentType, message.content, function (err, result) {
            if (err) {
                throw {httpStatus: 500, message: 'Save private message failed.'};
            }

            res.writeHead(200);

            // 更新自身的聊天统计信息
            msgStats.updateP2PChatInfo(message.from,
                message.to,
                message.from,
                CHANNELS.Private,
                message.content,
                function (err, result) {
                    if (err) log.error(err);
                });

            // 更新对端的聊天统计信息
            msgStats.updateP2PChatInfo(message.to,
                message.from,
                message.from,
                message.contentType,
                message.content,
                function (err, result) {
                    if (err) log.error(err);
                });

            // 获取对方状态，即对端的系统平台，token等信息，并推送通知消息给对端
            user.getUserStatus(message.to, function (err, result) {
                if (err) {
                    log.error('Get users status failed： ' + message.to);

                    throw {httpStatus: 500, message: 'Find target users failed.'};
                }

                // 构建通知消息
                var title = '新消息';
                var content = '';
                if (message.contentType == 1) {
                    content = message.content;
                } else if (message.contentType == 2) {
                    content = '[图片]';
                } else if (message.contentType == 3) {
                    content = '[语音]';
                } else {
                    content = '接收到一条新消息';
                }

                var pushable = false;
                var target;
                if (result.length > 0) {
                    target = result[0];
                    if (target.is_online) {
                        pushable = true;
                    }
                }

                var notifyMessage = JSON.stringify({type: 'p2p_msg', from: message.from});

                // 保存通知消息到数据库中并根据用户在线状态推送此消息
                notifyMsg.saveNotifyMessage(message.to,
                    message.contentType,
                    title,
                    content,
                    notifyMessage,
                    pushable,
                    function (err, result) {
                        if (err) {
                            log.error('Save private notify message failed, ', err);
                        } else {
                            if (pushable === true) {
                                if (data.platform === PLATFORMS.iOS) {
                                    getui.pushAPN(message.to,
                                        data.token,
                                        message.contentType,
                                        title,
                                        content,
                                        notifyMessage,
                                        function (err, result) {
                                            if (err != null) {
                                                console.log(err);
                                            } else {
                                                console.log(result);
                                            }
                                        });
                                } else if (data.platform === PLATFORMS.Android) {
                                    getui.pushAndroid(data.client_id,
                                        message.contentType,
                                        title,
                                        content,
                                        notifyMessage,
                                        data.status,
                                        function (err, result) {
                                            if (err != null) {
                                                console.log(err);
                                            } else {
                                                console.log(result);
                                            }
                                        });
                                }
                            }
                        }
                    });
            });
        });
    });

    res.writeHead(200);
});

/**
 * 处理讨论组消息。
 *
 * 消息格式：
 *  {
 *      from: "sand",                       // 发送者id
 *      at: "Rose",                         // @人
 *      group: "DiscussionGroupId",         // 所在组
 *      groupType: "1 or 2",                // 组类型：行政团队或讨论组
 *      contentType: "1,2,3",               // 内容类型
 *      content: "The patient mess me up"   // 内容
 *  }
 *
 * @param message
 */
router.post(APIv1.Chats.GM, function (req, res) {
    // 检查消息体及消息格式是否正确
    var message = req.body;
    if (!objectUtil.isJsonObject(message)) {
        throw {httpStatus: 406, message: 'Problems parsing JSON.'};
    }

    // 字段判断
    if (message.from == null || message.group == null || message.contentType == null || message.content == null) {
        throw {httpStatus: 406, message: 'Missing fields.'}
    }

    // 消息处理
    group.isGroupMember(message.group, message.groupType, message.from, function (err, result) {
        if (err) {
            throw {httpStatus: 500, message: 'Check group member failed.'};
        }

        if (result.length == 0) {
            throw {httpStatus: 404, message: 'Group member not found. Member id: ' + message.from};
        }

        // 保存群组消息
        groupMsg.saveMessage(message.from, message.group, message.at, message.contentType, message.content, function (err, result) {
            if (err) {
                throw {httpStatus: 500, message: 'Save group message failed.'};
            }

            res.writeHead(200);

            // 统计信息
            msgStats.updateGroupChatInfo(message.from,
                message.group,
                message.from,
                0,
                message.contentType,
                message.content,
                false,
                function (err, result) {
                    if (err) log.error(err);
                });

            // 推送通知消息给群组成员
            group.getMembers(message.group, message.groupType, function (err, result) {
                if (err) {
                    log.error('Get group members failed: ', err);
                    throw {httpStatus: 500, message: 'Get group members failed.'};
                }

                if (result.length == 0) {
                    log.warn('No members in group ', message.group);
                    return;
                }

                // 逐个推送通知
                for (var nIndex = 0; nIndex < result.length; nIndex++) {
                    var member = result[nIndex];
                    if (member.member_code === message.from) continue;

                    user.getUserStatus(member.member_code, function (err, result) {
                        if (err) {
                            console.error('Get group member status failed: ', err);
                            return;
                        }

                        var title = '';
                        var content = '';
                        if (message.contentType === 1) {
                            title = '群组消息';
                            content = message.content;
                        } else if (message.contentType === 2) {
                            title = '群组消息';
                            content = '[图片]';
                        } else if (message.contentType === 3) {
                            title = '群组消息';
                            content = '[语音]';
                        } else {
                            title = '群组消息';
                            content = '接收到一条新消息';
                        }

                        var pushable = result.length > 0 && result[0].is_online === 1;
                        var notifyMessage = JSON.stringify({type: 'group_msg', groupId: message.group});

                        // 发送并保存通知到数据库中
                        if (pushable) {
                            if (data.platform === PLATFORMS.iOS) {
                                getui.pushAPN(member.member_code,
                                    data.token,
                                    message.contentType,
                                    title,
                                    content,
                                    notifyMessage,
                                    function (err, result) {
                                        err != null ? console.error(err) : console.log(result);

                                        // 这段代码重复
                                        notifyMsg.saveNotifyMessage(member.member_code,
                                            message.contentType,
                                            title,
                                            content,
                                            notifyMessage,
                                            err != null ? 0 : 1,
                                            function (err, result) {
                                                if (err) {
                                                    console.log('Save group notify message failed: ', err);
                                                } else {

                                                }
                                            });
                                    });
                            } else if (data.platform === PLATFORMS.Android) {
                                getui.pushAndroid(data.client_id,
                                    message.contentType,
                                    title,
                                    content,
                                    notifyMessage,
                                    data.status,
                                    function (err, result) {
                                        err != null ? console.error(err) : console.log(result);

                                        // 这段代码重复
                                        notifyMsg.saveNotifyMessage(member.member_code,
                                            message.contentType,
                                            title,
                                            content,
                                            notifyMessage,
                                            err != null ? 0 : 1,
                                            function (err, result) {
                                                if (err) {
                                                    console.log('Save group notify message failed: ', err);
                                                } else {

                                                }
                                            });
                                    });
                            }
                        } else {
                            // 这段代码重复
                            notifyMsg.saveNotifyMessage(member.member_code,
                                message.contentType,
                                title,
                                content,
                                notifyMessage,
                                0,
                                function (err, result) {
                                    if (err) {
                                        console.log('Save group notify message failed: ', err);
                                    } else {

                                    }
                                });
                        }
                    });

                    // 统计信息
                    var at = 0;
                    if (message.at == member.member_code) {
                        at = 1;
                    }

                    msgStats.updateGroupChatInfo(member.member_code,
                        message.group,
                        message.from,
                        at,
                        message.contentType,
                        message.content,
                        true,
                        function (err, result) {
                            if (err) console.log(err);
                        });
                }
            });
        });
    });
});

//--------------------------------------------------------------//
//----------------------------消息获取----------------------------//
//--------------------------------------------------------------//

/**
 * 按时间倒序获取私信。
 *
 * 参数：
 *  user_id 必须，医生ID
 *  peer_id 必须，对方医生ID
 *  message_start_id 可选，消息的起始ID，如果为空从最新的一条开始获取
 *  count 可选，消息数量，如果不指定、小于零或大于50，默认为50条
 *
 * 请求URL：
 *  /pm?user_id=sand&peer_id=Rose&message_start_id=0&count=20
 */
router.get(APIv1.Chats.PM, function (req, res)  {
    var userId = req.query.user_id;
    var peerId = req.query.peer_id;
    var msgStartId = req.query.message_start_id === null ? MAX_INT : parseInt(req.query.message_start_id);
    var count = req.query.count == null ? 50 : parsetInt(req.query.count);

    if (userId == null) {
        throw {httpStatus: 400, message: "Missing field."};
    }

    privateMsg.findMessage(userId, peerId, msgStartId, count, function (err, rows) {
        if (err) {
            log.error("Get private message failed, ", err);

            throw {httpStatus: 500, message: "Get private messages failed."};
        }

        var messages = {};
        messages.startId = msgStartId;
        messages.count = rows.length;
        messages.records = [];
        for (var nIndex = 0; nIndex < rows.length; nIndex++) {
            rows[nIndex].timestamp = Date.parse(new Date(rows[nIndex].timestamp));
            messages.records.notifyMsg(rows[nIndex]);
        }

        res.writeHead(200);
        res.write(messages);

        // 清空统计信息
        msgStats.clearP2PChatInfo(userId, peerId, function (err, result) {
            if (err) console.log(err);
        });
    });
});

/**
 * 按时间倒序获取群消息。
 *
 * 参数：
 *  group_id 必须，组ID
 *  user_id 必须，医生ID
 *  message_start_id 可选，消息的起始ID，如果为空从最新的一条开始获取
 *  count 可选，消息数量，如果不指定、小于零或大于50，默认为50条
 *
 * 请求URL：
 *  /gm?user_id=sand&group_id=discussionGroup&message_start_id=0&count=20
 */
router.get(APIv1.Chats.GM, function (req, res)  {
    var groupId = req.query.group_id;
    var userId = req.query.user_id;
    var msgStartId = req.query.message_start_id === null ? MAX_INT : parseInt(req.query.message_start_id);
    var count = req.query.count == null ? 50 : parsetInt(req.query.count);

    if (groupId == null) {
        throw {httpStatus: 400, message: "Missing field."};
    }

    groupMsg.findMessage(groupId, msgStartId, count, function (err, rows) {
        if (err) {
            console.log('Get group message failed: ', err);

            throw {httpStatus: 500, message: 'Get group message failed.'};
        }

        var messages = {};
        messages.startId = msgStartId;
        messages.count = rows.length;
        messages.records = [];
        for (var nIndex = 0; nIndex < rows.length; nIndex++) {
            rows[nIndex].timestamp = Date.parse(new Date(rows[nIndex].timestamp));
            messages.records.push(rows[nIndex]);
        }

        res.writeHead(200);
        res.write(messages);

        // 清空统计信息
        msgStats.clearGroupChatInfo(req.query.uid,
            userId,
            function (err, result) {
                if (err) console.log(err);
            });
    });
});

//--------------------------------------------------------------//
//----------------------------消息统计----------------------------//
//--------------------------------------------------------------//

/**
 * 获取参与的聊天列表，包括：点对点，@我，参与的讨论组。
 *
 * 请求URL：
 *  /api/v1/chats?user_id=sand
 */
router.get(APIv1.Chats.List, function (req, res) {
    var userId = req.query.user_id;
    if (userId === null) {
        throw {httpStatus: 406, message: 'Missing fields.'};
    }

    msgStats.getChatList(userId, function (err, rows) {
        if (err) {
            log.error('Get users chat list failed: ', err);

            res.status(500).send({message: 'Get users chat list failed.'});
            return;
        }

        if (rows.length == 0) {
            res.status(200).send([]);
            return;
        }

        var chats = new Array(rows.length);
        for (var i = 0; i < rows.length; i++) {
            var row = rows[i];
            var element = {
                userId: row.uid,
                peerId: row.from_uid,
                groupId: row.from_gid === undefined ? "" : row.from_gid,
                atMe: row.at_me === 1,
                messageType: row.msg_type,
                lastContentType: row.last_content_type,
                lastContent: row.last_content,
                newMessageCount: row.new_msg_count,
                timestamp: Date.parse(new Date(row.timestamp))
            };

            chats[i] = element;
        }

        res.status(200).send(chats);
    });
});

/**
 * 获取所有群组未读消息总数。
 *
 * 请求URL：
 *  /gm/unread_count?user_id=sand
 *
 * 参数：
 *  user_id：医生ID
 */
router.get(APIv1.Chats.GMUnreadCount, function (req, res) {
    var userId = req.query.user_id;
    if (userId === null) {
        throw {httpStatus: 406, message: 'Missing fields.'};
    }

    msgStats.getGroupChatAllUnRead(userId, function (err, result) {
        if (err) {
            console.log('Get all unread messages failed: ', err);

            res.status(500).send({message: 'Get all unread messages failed.'});
        }

        var data = {
            userId: userId,
            messageType: 2,
            newMessageCount: 0
        };

        if (result.length > 0) {
            for (var index = 0; index < result.length; index++) {
                data.newMessageCount += result[index].new_msg_count;
            }
        }

        res.send(data);
    });
});

/**
 * 获取特定群组消息统计情况。
 *
 * /gm/statistic?group_id=GGG&&user_id=sand
 *
 * 参数：
 *  user_id：信息所有者id
 *  group_id：群组id
 */
router.get(APIv1.Chats.GMStats, function (req, res) {
    var userId = req.query.user_id;
    var groupId = req.query.group_id;
    if (userId === null || groupId === null) {
        throw {httpStatus: 406, message: 'Miss fields.'};
    }

    msgStats.getGroupChatInfo(userId, groupId, function (err, result) {
        if (err) {
            console.log('Get group stats failed: ', err);

            throw {httpStatus: 500, message: 'Get group stats failed.'};
        }

        if (result.length == 0) {
            var data = {
                "userId": userId,
                "from": "",
                "groupId": groupId,
                "atMe": 0,
                "lastContentType": 1,
                "lastContent": "",
                "newMessageCount": 0,
                "timestamp": 0
            };

            res.write(data);
            return;
        }

        result[0].timestamp = Date.parse(new Date(result[0].timestamp));
        res.write(result[0]);
    });
});

/**
 * 获取与某人的私信统计。
 *
 * /pm/statistic?user_id=sand&&peer_id=rose
 *
 * 参数：
 *  user_id：信息所有者id
 *  peer_id：聊天对端id
 */
router.get(APIv1.Chats.PMStats, function (req, res) {
    var userId =  req.query.user_id;
    var peerId =  req.query.peer_id;
    if (userId == null || peerId == null) {
        throw {httpStatus: 406, message: "Missing fields."};
    }

    msgStats.getP2PChatInfo(userId, peerId, function (err, result) {
        if (err) {
            console.log("Get private messages stats failed: ", err);

            throw {httpStatus: 500, message: "Get private messages stats failed."};
        }

        if (result.length == 0) {
            var data = {
                "userId": userId,
                "peerId": peerId,
                "lastCContentType": 1,
                "lastContent": "",
                "newMessageCount": 0,
                "timestamp": 0
            };
            res.write(data);
            return;
        }

        result[0].timestamp = Date.parse(new Date(result[0].timestamp));
        res.write(result[0]);
    });
});

/**
 * 获取所有未读私信总数。
 *
 * /pm/unread_count?user_id=sand
 *
 * 参数：
 * uid：信息所有者id
 */
router.get(APIv1.Chats.PMUnreadCount, function (req, res) {
    var userId = req.query.user_id;

    msgStats.getP2PChatAllUnRead(userId, function (err, result) {
        if (err) {
            console.log("Get unread private message count failed: ", err);

            res.status(500).send({message: "Get unread private message count failed."});
        }

        var data = {userId: userId, messageType: 1, newMessageCount: 0};
        if (result.length > 0) {
            for (var i = 0; i < result.length; i++) {
                data.newMessageCount += result[i].new_msg_count;
            }
        }

        res.send(data);
    });
});

/**
 * 所有聊天消息未读数。
 *
 * 请求URL：
 *  /chats/unread_count?user_id=sand
 *
 * 参数：
 *  user_id：信息所有者id
 */
router.get(APIv1.Chats.UnreadMsgCount, function (req, res) {
    var userId = req.query.user_id;
    if (userId === null) {
        throw {httpStatus: 406, message: "Missing fields."};
    }

    msgStats.getChatAllUnRead(userId, function (err, result) {
        if (err) {
            console.error("Get all unread message count failed: ", err);

            res.status(500).send({message: "Get all unread message count failed."});
        }

        var data = {userId: userId, messageType: 0, newMessageCount: 0};
        if (result.length > 0) {
            for (var index = 0; index < result.length; index++) {
                data.newMessageCount += result[index].new_msg_count;
            }
        }

        res.send(data);
    });
});

module.exports = router;
